@@ -1,61 +1,55 @@
-#[rustfmt::skip]
-#[cfg(not(any(target_os = "macos", windows)))]
-use {
-    std::sync::atomic::AtomicBool,
-    std::sync::Arc,
-
-    glutin::platform::unix::{WindowBuilderExtUnix, WindowExtUnix},
-};
-
 #[rustfmt::skip]
 #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
 use {
     wayland_client::protocol::wl_surface::WlSurface,
     wayland_client::{Attached, EventQueue, Proxy},
-    glutin::platform::unix::EventLoopWindowTargetExtUnix,
-    glutin::window::Theme,
+    winit::platform::wayland::{EventLoopWindowTargetExtWayland, WindowExtWayland},
 };
 
+#[cfg(all(not(feature = "x11"), not(any(target_os = "macos", windows))))]
+use winit::platform::wayland::WindowBuilderExtWayland;
+
 #[rustfmt::skip]
 #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
 use {
     std::io::Cursor,
 
+    winit::platform::x11::{WindowExtX11, WindowBuilderExtX11},
+    glutin::platform::x11::X11VisualInfo,
     x11_dl::xlib::{Display as XDisplay, PropModeReplace, XErrorEvent, Xlib},
-    glutin::window::Icon,
+    winit::window::Icon,
     png::Decoder,
 };
 
 use std::fmt::{self, Display, Formatter};
-use std::ops::{Deref, DerefMut};
-use std::sync::atomic::{AtomicU8, Ordering};
+use std::sync::atomic::AtomicBool;
+use std::sync::Arc;
 
-use bitflags::bitflags;
 #[cfg(target_os = "macos")]
-use cocoa::base::{id, NO, YES};
-use glutin::dpi::{PhysicalPosition, PhysicalSize};
-use glutin::event_loop::EventLoopWindowTarget;
-#[cfg(target_os = "macos")]
-use glutin::platform::macos::{WindowBuilderExtMacOS, WindowExtMacOS};
-#[cfg(windows)]
-use glutin::platform::windows::IconExtWindows;
-use glutin::window::{
-    CursorIcon, Fullscreen, UserAttentionType, Window as GlutinWindow, WindowBuilder, WindowId,
+use {
+    cocoa::appkit::NSColorSpace,
+    cocoa::base::{id, nil, NO, YES},
+    objc::{msg_send, sel, sel_impl},
+    winit::platform::macos::{OptionAsAlt, WindowBuilderExtMacOS, WindowExtMacOS},
 };
-use glutin::{self, ContextBuilder, PossiblyCurrent, Rect, WindowedContext};
-#[cfg(target_os = "macos")]
-use objc::{msg_send, sel, sel_impl};
-#[cfg(target_os = "macos")]
+
 use raw_window_handle::{HasRawWindowHandle, RawWindowHandle};
+
+use winit::dpi::{PhysicalPosition, PhysicalSize};
+use winit::event_loop::EventLoopWindowTarget;
+use winit::monitor::MonitorHandle;
 #[cfg(windows)]
-use winapi::shared::minwindef::WORD;
+use winit::platform::windows::IconExtWindows;
+use winit::window::{
+    CursorIcon, Fullscreen, ImePurpose, UserAttentionType, Window as WinitWindow, WindowBuilder,
+    WindowId,
+};
 
 use alacritty_terminal::index::Point;
 
 use crate::config::window::{Decorations, Identity, WindowConfig};
 use crate::config::UiConfig;
 use crate::display::SizeInfo;
-use crate::gl;
 
 /// Window icon for `_NET_WM_ICON` property.
 #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
@@ -63,30 +57,16 @@ static WINDOW_ICON: &[u8] = include_bytes!("../../extra/logo/compat/alacritty-te
 
 /// This should match the definition of IDI_ICON from `alacritty.rc`.
 #[cfg(windows)]
-const IDI_ICON: WORD = 0x101;
-
-/// Context creation flags from probing config.
-static GL_CONTEXT_CREATION_FLAGS: AtomicU8 = AtomicU8::new(GlContextFlags::SRGB.bits);
-
-bitflags! {
-    pub struct GlContextFlags: u8 {
-        const EMPTY      = 0b000000000;
-        const SRGB       = 0b0000_0001;
-        const DEEP_COLOR = 0b0000_0010;
-    }
-}
+const IDI_ICON: u16 = 0x101;
 
 /// Window errors.
 #[derive(Debug)]
 pub enum Error {
     /// Error creating the window.
-    ContextCreation(glutin::CreationError),
+    WindowCreation(winit::error::OsError),
 
     /// Error dealing with fonts.
     Font(crossfont::Error),
-
-    /// Error manipulating the rendering context.
-    Context(glutin::ContextError),
 }
 
 /// Result of fallible operations concerning a Window.
@@ -95,8 +75,7 @@ type Result<T> = std::result::Result<T, Error>;
 impl std::error::Error for Error {
     fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
         match self {
-            Error::ContextCreation(err) => err.source(),
-            Error::Context(err) => err.source(),
+            Error::WindowCreation(err) => err.source(),
             Error::Font(err) => err.source(),
         }
     }
@@ -105,22 +84,15 @@ impl std::error::Error for Error {
 impl Display for Error {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         match self {
-            Error::ContextCreation(err) => write!(f, "Error creating GL context; {}", err),
-            Error::Context(err) => write!(f, "Error operating on render context; {}", err),
+            Error::WindowCreation(err) => write!(f, "Error creating GL context; {}", err),
             Error::Font(err) => err.fmt(f),
         }
     }
 }
 
-impl From<glutin::CreationError> for Error {
-    fn from(val: glutin::CreationError) -> Self {
-        Error::ContextCreation(val)
-    }
-}
-
-impl From<glutin::ContextError> for Error {
-    fn from(val: glutin::ContextError) -> Self {
-        Error::Context(val)
+impl From<winit::error::OsError> for Error {
+    fn from(val: winit::error::OsError) -> Self {
+        Error::WindowCreation(val)
     }
 }
 
@@ -130,41 +102,12 @@ impl From<crossfont::Error> for Error {
     }
 }
 
-fn create_gl_window<E>(
-    mut window: WindowBuilder,
-    event_loop: &EventLoopWindowTarget<E>,
-    flags: GlContextFlags,
-    vsync: bool,
-    dimensions: Option<PhysicalSize<u32>>,
-) -> Result<WindowedContext<PossiblyCurrent>> {
-    if let Some(dimensions) = dimensions {
-        window = window.with_inner_size(dimensions);
-    }
-
-    let mut windowed_context_builder = ContextBuilder::new()
-        .with_srgb(flags.contains(GlContextFlags::SRGB))
-        .with_vsync(vsync)
-        .with_hardware_acceleration(None);
-
-    if flags.contains(GlContextFlags::DEEP_COLOR) {
-        windowed_context_builder = windowed_context_builder.with_pixel_format(30, 2);
-    }
-
-    let windowed_context = windowed_context_builder.build_windowed(window, event_loop)?;
-
-    // Make the context current so OpenGL operations can run.
-    let windowed_context = unsafe { windowed_context.make_current().map_err(|(_, err)| err)? };
-
-    Ok(windowed_context)
-}
-
 /// A window which can be used for displaying the terminal.
 ///
 /// Wraps the underlying windowing library to provide a stable API in Alacritty.
 pub struct Window {
-    /// Flag tracking frame redraw requests from Wayland compositor.
-    #[cfg(not(any(target_os = "macos", windows)))]
-    pub should_draw: Arc<AtomicBool>,
+    /// Flag tracking that we have a frame we can draw.
+    pub has_frame: Arc<AtomicBool>,
 
     /// Attached Wayland surface to request new frame events.
     #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
@@ -173,10 +116,11 @@ pub struct Window {
     /// Cached scale factor for quickly scaling pixel sizes.
     pub scale_factor: f64,
 
+    window: WinitWindow,
+
     /// Current window title.
     title: String,
 
-    windowed_context: Replaceable<WindowedContext<PossiblyCurrent>>,
     current_mouse_cursor: CursorIcon,
     mouse_visible: bool,
 }
@@ -189,110 +133,112 @@ impl Window {
         event_loop: &EventLoopWindowTarget<E>,
         config: &UiConfig,
         identity: &Identity,
-        size: Option<PhysicalSize<u32>>,
         #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
         wayland_event_queue: Option<&EventQueue>,
+        #[rustfmt::skip]
+        #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
+        x11_visual: Option<X11VisualInfo>,
     ) -> Result<Window> {
         let identity = identity.clone();
-        let mut window_builder = Window::get_platform_window(&identity, &config.window);
+        let mut window_builder = Window::get_platform_window(
+            &identity,
+            &config.window,
+            #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
+            x11_visual,
+        );
 
         if let Some(position) = config.window.position {
             window_builder = window_builder
                 .with_position(PhysicalPosition::<i32>::from((position.x, position.y)));
         }
 
+        let window = window_builder
+            .with_title(&identity.title)
+            .with_theme(config.window.decorations_theme_variant)
+            .with_visible(false)
+            .with_transparent(true)
+            .with_maximized(config.window.maximized())
+            .with_fullscreen(config.window.fullscreen())
+            .build(event_loop)?;
+
         // Check if we're running Wayland to disable vsync.
         #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
         let is_wayland = event_loop.is_wayland();
-        #[cfg(any(not(feature = "wayland"), target_os = "macos", windows))]
+        #[cfg(all(not(feature = "wayland"), not(any(target_os = "macos", windows))))]
         let is_wayland = false;
 
-        let mut windowed_context = None;
-        let current_flags =
-            GlContextFlags::from_bits_truncate(GL_CONTEXT_CREATION_FLAGS.load(Ordering::Relaxed));
-        for flags in [
-            current_flags,
-            GlContextFlags::EMPTY,
-            GlContextFlags::SRGB | GlContextFlags::DEEP_COLOR,
-            GlContextFlags::DEEP_COLOR,
-        ] {
-            windowed_context = Some(create_gl_window(
-                window_builder.clone(),
-                event_loop,
-                flags,
-                !is_wayland,
-                size,
-            ));
-            if windowed_context.as_ref().unwrap().is_ok() {
-                GL_CONTEXT_CREATION_FLAGS.store(flags.bits, Ordering::Relaxed);
-                break;
-            }
-        }
-        let windowed_context = windowed_context.unwrap()?;
-
         // Text cursor.
         let current_mouse_cursor = CursorIcon::Text;
-        windowed_context.window().set_cursor_icon(current_mouse_cursor);
+        window.set_cursor_icon(current_mouse_cursor);
 
         // Enable IME.
-        windowed_context.window().set_ime_allowed(true);
+        window.set_ime_allowed(true);
+        window.set_ime_purpose(ImePurpose::Terminal);
 
-        // Set OpenGL symbol loader. This call MUST be after window.make_current on windows.
-        gl::load_with(|symbol| windowed_context.get_proc_address(symbol) as *const _);
+        // Set initial transparency hint.
+        window.set_transparent(config.window_opacity() < 1.);
+
+        #[cfg(target_os = "macos")]
+        use_srgb_color_space(&window);
 
         #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
         if !is_wayland {
             // On X11, embed the window inside another if the parent ID has been set.
             if let Some(parent_window_id) = config.window.embed {
-                x_embed_window(windowed_context.window(), parent_window_id);
+                x_embed_window(&window, parent_window_id);
             }
         }
 
         #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
         let wayland_surface = if is_wayland {
             // Attach surface to Alacritty's internal wayland queue to handle frame callbacks.
-            let surface = windowed_context.window().wayland_surface().unwrap();
+            let surface = window.wayland_surface().unwrap();
             let proxy: Proxy<WlSurface> = unsafe { Proxy::from_c_ptr(surface as _) };
             Some(proxy.attach(wayland_event_queue.as_ref().unwrap().token()))
         } else {
             None
         };
 
-        let scale_factor = windowed_context.window().scale_factor();
+        let scale_factor = window.scale_factor();
+        log::info!("Window scale factor: {}", scale_factor);
 
         Ok(Self {
             current_mouse_cursor,
             mouse_visible: true,
-            windowed_context: Replaceable::new(windowed_context),
+            window,
             title: identity.title,
-            #[cfg(not(any(target_os = "macos", windows)))]
-            should_draw: Arc::new(AtomicBool::new(true)),
+            has_frame: Arc::new(AtomicBool::new(true)),
             #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
             wayland_surface,
             scale_factor,
         })
     }
 
+    #[inline]
+    pub fn raw_window_handle(&self) -> RawWindowHandle {
+        self.window.raw_window_handle()
+    }
+
     #[inline]
     pub fn set_inner_size(&self, size: PhysicalSize<u32>) {
-        self.window().set_inner_size(size);
+        self.window.set_inner_size(size);
     }
 
     #[inline]
     pub fn inner_size(&self) -> PhysicalSize<u32> {
-        self.window().inner_size()
+        self.window.inner_size()
     }
 
     #[inline]
     pub fn set_visible(&self, visibility: bool) {
-        self.window().set_visible(visibility);
+        self.window.set_visible(visibility);
     }
 
     /// Set the window title.
     #[inline]
     pub fn set_title(&mut self, title: String) {
         self.title = title;
-        self.window().set_title(&self.title);
+        self.window.set_title(&self.title);
     }
 
     /// Get the window title.
@@ -303,14 +249,14 @@ impl Window {
 
     #[inline]
     pub fn request_redraw(&self) {
-        self.window().request_redraw();
+        self.window.request_redraw();
     }
 
     #[inline]
     pub fn set_mouse_cursor(&mut self, cursor: CursorIcon) {
         if cursor != self.current_mouse_cursor {
             self.current_mouse_cursor = cursor;
-            self.window().set_cursor_icon(cursor);
+            self.window.set_cursor_icon(cursor);
         }
     }
 
@@ -318,12 +264,18 @@ impl Window {
     pub fn set_mouse_visible(&mut self, visible: bool) {
         if visible != self.mouse_visible {
             self.mouse_visible = visible;
-            self.window().set_cursor_visible(visible);
+            self.window.set_cursor_visible(visible);
         }
     }
 
     #[cfg(not(any(target_os = "macos", windows)))]
-    pub fn get_platform_window(identity: &Identity, window_config: &WindowConfig) -> WindowBuilder {
+    pub fn get_platform_window(
+        identity: &Identity,
+        window_config: &WindowConfig,
+        #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))] x11_visual: Option<
+            X11VisualInfo,
+        >,
+    ) -> WindowBuilder {
         #[cfg(feature = "x11")]
         let icon = {
             let mut decoder = Decoder::new(Cursor::new(WINDOW_ICON));
@@ -336,55 +288,33 @@ impl Window {
         };
 
         let builder = WindowBuilder::new()
-            .with_title(&identity.title)
             .with_name(&identity.class.general, &identity.class.instance)
-            .with_visible(false)
-            .with_transparent(true)
-            .with_decorations(window_config.decorations != Decorations::None)
-            .with_maximized(window_config.maximized())
-            .with_fullscreen(window_config.fullscreen());
+            .with_decorations(window_config.decorations != Decorations::None);
 
         #[cfg(feature = "x11")]
         let builder = builder.with_window_icon(Some(icon));
 
         #[cfg(feature = "x11")]
-        let builder = match window_config.decorations_theme_variant() {
-            Some(val) => builder.with_gtk_theme_variant(val.to_string()),
+        let builder = match x11_visual {
+            Some(visual) => builder.with_x11_visual(visual.into_raw()),
             None => builder,
         };
 
-        #[cfg(feature = "wayland")]
-        let builder = match window_config.decorations_theme_variant() {
-            Some("light") => builder.with_wayland_csd_theme(Theme::Light),
-            // Prefer dark theme by default, since default alacritty theme is dark.
-            _ => builder.with_wayland_csd_theme(Theme::Dark),
-        };
-
         builder
     }
 
     #[cfg(windows)]
-    pub fn get_platform_window(identity: &Identity, window_config: &WindowConfig) -> WindowBuilder {
-        let icon = glutin::window::Icon::from_resource(IDI_ICON, None);
+    pub fn get_platform_window(_: &Identity, window_config: &WindowConfig) -> WindowBuilder {
+        let icon = winit::window::Icon::from_resource(IDI_ICON, None);
 
         WindowBuilder::new()
-            .with_title(&identity.title)
-            .with_visible(false)
             .with_decorations(window_config.decorations != Decorations::None)
-            .with_transparent(true)
-            .with_maximized(window_config.maximized())
-            .with_fullscreen(window_config.fullscreen())
             .with_window_icon(icon.ok())
     }
 
     #[cfg(target_os = "macos")]
-    pub fn get_platform_window(identity: &Identity, window_config: &WindowConfig) -> WindowBuilder {
-        let window = WindowBuilder::new()
-            .with_title(&identity.title)
-            .with_visible(false)
-            .with_transparent(true)
-            .with_maximized(window_config.maximized())
-            .with_fullscreen(window_config.fullscreen());
+    pub fn get_platform_window(_: &Identity, window_config: &WindowConfig) -> WindowBuilder {
+        let window = WindowBuilder::new().with_option_as_alt(window_config.option_as_alt);
 
         match window_config.decorations {
             Decorations::Full => window,
@@ -404,47 +334,64 @@ impl Window {
     pub fn set_urgent(&self, is_urgent: bool) {
         let attention = if is_urgent { Some(UserAttentionType::Critical) } else { None };
 
-        self.window().request_user_attention(attention);
+        self.window.request_user_attention(attention);
     }
 
     pub fn id(&self) -> WindowId {
-        self.window().id()
+        self.window.id()
+    }
+
+    pub fn set_transparent(&self, transparent: bool) {
+        self.window.set_transparent(transparent);
     }
 
     pub fn set_maximized(&self, maximized: bool) {
-        self.window().set_maximized(maximized);
+        self.window.set_maximized(maximized);
     }
 
     pub fn set_minimized(&self, minimized: bool) {
-        self.window().set_minimized(minimized);
+        self.window.set_minimized(minimized);
+    }
+
+    pub fn set_resize_increments(&self, increments: PhysicalSize<f32>) {
+        self.window.set_resize_increments(Some(increments));
     }
 
     /// Toggle the window's fullscreen state.
     pub fn toggle_fullscreen(&self) {
-        self.set_fullscreen(self.window().fullscreen().is_none());
+        self.set_fullscreen(self.window.fullscreen().is_none());
     }
 
     /// Toggle the window's maximized state.
     pub fn toggle_maximized(&self) {
-        self.set_maximized(!self.window().is_maximized());
+        self.set_maximized(!self.window.is_maximized());
     }
 
     #[cfg(target_os = "macos")]
     pub fn toggle_simple_fullscreen(&self) {
-        self.set_simple_fullscreen(!self.window().simple_fullscreen());
+        self.set_simple_fullscreen(!self.window.simple_fullscreen());
+    }
+
+    #[cfg(target_os = "macos")]
+    pub fn set_option_as_alt(&self, option_as_alt: OptionAsAlt) {
+        self.window.set_option_as_alt(option_as_alt);
     }
 
     pub fn set_fullscreen(&self, fullscreen: bool) {
         if fullscreen {
-            self.window().set_fullscreen(Some(Fullscreen::Borderless(None)));
+            self.window.set_fullscreen(Some(Fullscreen::Borderless(None)));
         } else {
-            self.window().set_fullscreen(None);
+            self.window.set_fullscreen(None);
         }
     }
 
+    pub fn current_monitor(&self) -> Option<MonitorHandle> {
+        self.window.current_monitor()
+    }
+
     #[cfg(target_os = "macos")]
     pub fn set_simple_fullscreen(&self, simple_fullscreen: bool) {
-        self.window().set_simple_fullscreen(simple_fullscreen);
+        self.window.set_simple_fullscreen(simple_fullscreen);
     }
 
     #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
@@ -453,7 +400,7 @@ impl Window {
     }
 
     pub fn set_ime_allowed(&self, allowed: bool) {
-        self.windowed_context.window().set_ime_allowed(allowed);
+        self.window.set_ime_allowed(allowed);
     }
 
     /// Adjust the IME editor position according to the new location of the cursor.
@@ -461,56 +408,7 @@ impl Window {
         let nspot_x = f64::from(size.padding_x() + point.column.0 as f32 * size.cell_width());
         let nspot_y = f64::from(size.padding_y() + (point.line + 1) as f32 * size.cell_height());
 
-        self.window().set_ime_position(PhysicalPosition::new(nspot_x, nspot_y));
-    }
-
-    pub fn swap_buffers(&self) {
-        self.windowed_context.swap_buffers().expect("swap buffers");
-    }
-
-    pub fn swap_buffers_with_damage(&self, damage: &[Rect]) {
-        self.windowed_context.swap_buffers_with_damage(damage).expect("swap buffes with damage");
-    }
-
-    #[cfg(any(target_os = "macos", windows))]
-    pub fn swap_buffers_with_damage_supported(&self) -> bool {
-        // Disable damage tracking on macOS/Windows since there's no observation of it working.
-        false
-    }
-
-    #[cfg(not(any(target_os = "macos", windows)))]
-    pub fn swap_buffers_with_damage_supported(&self) -> bool {
-        // On X11 damage tracking is behaving in unexpected ways on some NVIDIA systems. Since
-        // there's no compositor supporting it, damage tracking is disabled on X11.
-        //
-        // For more see https://github.com/alacritty/alacritty/issues/6051.
-        #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
-        if self.window().xlib_window().is_some() {
-            return false;
-        }
-
-        self.windowed_context.swap_buffers_with_damage_supported()
-    }
-
-    pub fn resize(&self, size: PhysicalSize<u32>) {
-        self.windowed_context.resize(size);
-    }
-
-    pub fn make_not_current(&mut self) {
-        if self.windowed_context.is_current() {
-            self.windowed_context.replace_with(|context| unsafe {
-                // We do ensure that context is current before any rendering operation due to multi
-                // window support, so we don't need extra "type aid" from glutin here.
-                context.make_not_current().expect("context swap").treat_as_current()
-            });
-        }
-    }
-
-    pub fn make_current(&mut self) {
-        if !self.windowed_context.is_current() {
-            self.windowed_context
-                .replace_with(|context| unsafe { context.make_current().expect("context swap") });
-        }
+        self.window.set_ime_position(PhysicalPosition::new(nspot_x, nspot_y));
     }
 
     /// Disable macOS window shadows.
@@ -518,7 +416,7 @@ impl Window {
     /// This prevents rendering artifacts from showing up when the window is transparent.
     #[cfg(target_os = "macos")]
     pub fn set_has_shadow(&self, has_shadows: bool) {
-        let raw_window = match self.window().raw_window_handle() {
+        let raw_window = match self.raw_window_handle() {
             RawWindowHandle::AppKit(handle) => handle.ns_window as id,
             _ => return,
         };
@@ -528,14 +426,10 @@ impl Window {
             let _: id = msg_send![raw_window, setHasShadow: value];
         }
     }
-
-    fn window(&self) -> &GlutinWindow {
-        self.windowed_context.window()
-    }
 }
 
 #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
-fn x_embed_window(window: &GlutinWindow, parent_id: std::os::raw::c_ulong) {
+fn x_embed_window(window: &WinitWindow, parent_id: std::os::raw::c_ulong) {
     let (xlib_display, xlib_window) = match (window.xlib_display(), window.xlib_window()) {
         (Some(display), Some(window)) => (display, window),
         _ => return,
@@ -568,49 +462,20 @@ fn x_embed_window(window: &GlutinWindow, parent_id: std::os::raw::c_ulong) {
     }
 }
 
+#[cfg(target_os = "macos")]
+fn use_srgb_color_space(window: &WinitWindow) {
+    let raw_window = match window.raw_window_handle() {
+        RawWindowHandle::AppKit(handle) => handle.ns_window as id,
+        _ => return,
+    };
+
+    unsafe {
+        let _: () = msg_send![raw_window, setColorSpace: NSColorSpace::sRGBColorSpace(nil)];
+    }
+}
+
 #[cfg(all(feature = "x11", not(any(target_os = "macos", windows))))]
 unsafe extern "C" fn xembed_error_handler(_: *mut XDisplay, _: *mut XErrorEvent) -> i32 {
     log::error!("Could not embed into specified window.");
     std::process::exit(1);
 }
-
-/// Struct for safe in-place replacement.
-///
-/// This struct allows easily replacing struct fields that provide `self -> Self` methods in-place,
-/// without having to deal with constantly unwrapping the underlying [`Option`].
-struct Replaceable<T>(Option<T>);
-
-impl<T> Replaceable<T> {
-    pub fn new(inner: T) -> Self {
-        Self(Some(inner))
-    }
-
-    /// Replace the contents of the container.
-    pub fn replace_with<F: FnMut(T) -> T>(&mut self, f: F) {
-        self.0 = self.0.take().map(f);
-    }
-
-    /// Get immutable access to the wrapped value.
-    pub fn get(&self) -> &T {
-        self.0.as_ref().unwrap()
-    }
-
-    /// Get mutable access to the wrapped value.
-    pub fn get_mut(&mut self) -> &mut T {
-        self.0.as_mut().unwrap()
-    }
-}
-
-impl<T> Deref for Replaceable<T> {
-    type Target = T;
-
-    fn deref(&self) -> &Self::Target {
-        self.get()
-    }
-}
-
-impl<T> DerefMut for Replaceable<T> {
-    fn deref_mut(&mut self) -> &mut Self::Target {
-        self.get_mut()
-    }
-}
