@@ -37,8 +37,9 @@ pub fn builtin_glyph(
 }
 
 fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> RasterizedGlyph {
-    let height = (metrics.line_height as i32 + offset.y as i32) as usize;
-    let width = (metrics.average_advance as i32 + offset.x as i32) as usize;
+    // Ensure that width and height is at least one.
+    let height = (metrics.line_height as i32 + offset.y as i32).max(1) as usize;
+    let width = (metrics.average_advance as i32 + offset.x as i32).max(1) as usize;
     // Use one eight of the cell width, since this is used as a step size for block elemenets.
     let stroke_size = cmp::max((width as f32 / 8.).round() as usize, 1);
     let heavy_stroke_size = stroke_size * 2;
@@ -71,14 +72,14 @@ fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> Raster
             for stroke_size in 0..2 * stroke_size {
                 let stroke_size = stroke_size as f32 / 2.;
                 if character == '\u{2571}' || character == '\u{2573}' {
-                    let h = y_end - stroke_size as f32;
+                    let h = y_end - stroke_size;
                     let from_y = f_x(from_x, h);
                     let to_y = f_x(to_x, h);
                     canvas.draw_line(from_x, from_y, to_x, to_y);
                 }
                 if character == '\u{2572}' || character == '\u{2573}' {
-                    let from_y = g_x(from_x, stroke_size as f32);
-                    let to_y = g_x(to_x, stroke_size as f32);
+                    let from_y = g_x(from_x, stroke_size);
+                    let to_y = g_x(to_x, stroke_size);
                     canvas.draw_line(from_x, from_y, to_x, to_y);
                 }
             }
@@ -345,7 +346,7 @@ fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> Raster
             if character == '\u{256d}' || character == '\u{2570}' {
                 let center = canvas.x_center() as usize;
 
-                let extra_offset = if stroke_size % 2 == width % 2 { 0 } else { 1 };
+                let extra_offset = usize::from(stroke_size % 2 != width % 2);
 
                 let buffer = canvas.buffer_mut();
                 for y in 1..height {
@@ -363,7 +364,7 @@ fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> Raster
             if character == '\u{256d}' || character == '\u{256e}' {
                 let center = canvas.y_center() as usize;
 
-                let extra_offset = if stroke_size % 2 == height % 2 { 0 } else { 1 };
+                let extra_offset = usize::from(stroke_size % 2 != height % 2);
 
                 let buffer = canvas.buffer_mut();
                 if extra_offset != 0 {
@@ -417,12 +418,12 @@ fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> Raster
 
             // Ensure that resulted glyph will be visible and also round sizes instead of straight
             // flooring them.
-            rect_width = cmp::max(rect_width.round() as i32, 1) as f32;
-            rect_height = cmp::max(rect_height.round() as i32, 1) as f32;
+            rect_width = rect_width.round().max(1.);
+            rect_height = rect_height.round().max(1.);
 
             let x = match character {
                 '\u{2590}' => canvas.x_center(),
-                '\u{2595}' => width as f32 - rect_width,
+                '\u{2595}' => width - rect_width,
                 _ => 0.,
             };
 
@@ -441,27 +442,30 @@ fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> Raster
         },
         // Quadrants: '▖', '▗', '▘', '▙', '▚', '▛', '▜', '▝', '▞', '▟'.
         '\u{2596}'..='\u{259F}' => {
+            let x_center = canvas.x_center().round().max(1.);
+            let y_center = canvas.y_center().round().max(1.);
+
             let (w_second, h_second) = match character {
                 '\u{2598}' | '\u{2599}' | '\u{259a}' | '\u{259b}' | '\u{259c}' => {
-                    (canvas.x_center(), canvas.y_center())
+                    (x_center, y_center)
                 },
                 _ => (0., 0.),
             };
             let (w_first, h_first) = match character {
                 '\u{259b}' | '\u{259c}' | '\u{259d}' | '\u{259e}' | '\u{259f}' => {
-                    (canvas.x_center(), canvas.y_center())
+                    (x_center, y_center)
                 },
                 _ => (0., 0.),
             };
             let (w_third, h_third) = match character {
                 '\u{2596}' | '\u{2599}' | '\u{259b}' | '\u{259e}' | '\u{259f}' => {
-                    (canvas.x_center(), canvas.y_center())
+                    (x_center, y_center)
                 },
                 _ => (0., 0.),
             };
             let (w_fourth, h_fourth) = match character {
                 '\u{2597}' | '\u{2599}' | '\u{259a}' | '\u{259c}' | '\u{259f}' => {
-                    (canvas.x_center(), canvas.y_center())
+                    (x_center, y_center)
                 },
                 _ => (0., 0.),
             };
@@ -469,11 +473,11 @@ fn box_drawing(character: char, metrics: &Metrics, offset: &Delta<i8>) -> Raster
             // Second quadrant.
             canvas.draw_rect(0., 0., w_second, h_second, COLOR_FILL);
             // First quadrant.
-            canvas.draw_rect(canvas.x_center(), 0., w_first, h_first, COLOR_FILL);
+            canvas.draw_rect(x_center, 0., w_first, h_first, COLOR_FILL);
             // Third quadrant.
-            canvas.draw_rect(0., canvas.y_center(), w_third, h_third, COLOR_FILL);
+            canvas.draw_rect(0., y_center, w_third, h_third, COLOR_FILL);
             // Fourth quadrant.
-            canvas.draw_rect(canvas.x_center(), canvas.y_center(), w_fourth, h_fourth, COLOR_FILL);
+            canvas.draw_rect(x_center, y_center, w_fourth, h_fourth, COLOR_FILL);
         },
         _ => unreachable!(),
     }
@@ -592,13 +596,13 @@ impl Canvas {
     /// Draws a horizontal straight line from (`x`, `y`) of `size` with the given `stroke_size`.
     fn draw_h_line(&mut self, x: f32, y: f32, size: f32, stroke_size: usize) {
         let (start_y, end_y) = self.h_line_bounds(y, stroke_size);
-        self.draw_rect(x, start_y as f32, size, (end_y - start_y) as f32, COLOR_FILL);
+        self.draw_rect(x, start_y, size, end_y - start_y, COLOR_FILL);
     }
 
     /// Draws a vertical straight line from (`x`, `y`) of `size` with the given `stroke_size`.
     fn draw_v_line(&mut self, x: f32, y: f32, size: f32, stroke_size: usize) {
         let (start_x, end_x) = self.v_line_bounds(x, stroke_size);
-        self.draw_rect(start_x as f32, y, (end_x - start_x) as f32, size, COLOR_FILL);
+        self.draw_rect(start_x, y, end_x - start_x, size, COLOR_FILL);
     }
 
     /// Draws a rect from the (`x`, `y`) of the given `width` and `height` using `color`.
@@ -715,9 +719,8 @@ impl Canvas {
         let v_line_bounds = (v_line_bounds.0 as usize, v_line_bounds.1 as usize);
         let max_transparancy = 0.5;
 
-        for (radius_y, radius_x) in (h_line_bounds.0..h_line_bounds.1)
-            .into_iter()
-            .zip((v_line_bounds.0..v_line_bounds.1).into_iter())
+        for (radius_y, radius_x) in
+            (h_line_bounds.0..h_line_bounds.1).zip(v_line_bounds.0..v_line_bounds.1)
         {
             let radius_x = radius_x as f32;
             let radius_y = radius_y as f32;
@@ -750,7 +753,7 @@ impl Canvas {
 
                 let x_next = (x + 1.).clamp(0., v_line_bounds.1 as f32 - 1.);
                 let x = x.clamp(0., v_line_bounds.1 as f32 - 1.);
-                let y = y.clamp(0., radius_y as f32);
+                let y = y.clamp(0., radius_y);
 
                 self.put_pixel(x, y, color_1);
                 self.put_pixel(x_next, y, color_2);
